from django.test import TestCase
from django.urls import reverse

from django_marina.test import ExtendedClient


class ExtendedClientTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.url_echo = reverse("echo")

    def test_get(self):
        client = ExtendedClient()
        response = client.get(f"{self.url_echo}?foo=bar")
        self.assertContains(response, "foo")
        self.assertContains(response, "bar")

    def test_post(self):
        client = ExtendedClient()
        response = client.post(self.url_echo, data={"foo": "bar"})
        self.assertContains(response, "foo")
        self.assertContains(response, "bar")



from bs4 import BeautifulSoup
from django.test import Client
from django.utils.functional import SimpleLazyObject


def _get_soup(response):
    """Return soup for response."""
    if hasattr(response, "render") and callable(response.render) and not response.is_rendered:
        response.render()
    return BeautifulSoup(response.content, "html.parser")


class ExtendedClient(Client):
    """Client with added features."""

    USER_IGNORE = -1

    def request(self, **request):
        """Request a response from the server."""
        response = super().request(**request)
        # Inspired by http://blog.borys.info/2012/12/django-testing-with-beautifulsoup.html
        response.soup = SimpleLazyObject(lambda: _get_soup(response))
        return response

    def force_login(self, user, backend=None):
        """
        Force given user to login.

        When `user` is `ExtendedClient.USER_IGNORE`, this method does nothing.
        When `user` is `None`, this will force a logout.
        In all other cases, `user` will be logged in.
        """
        if user != self.USER_IGNORE:
            if user is None:
                self.logout()
            else:
                super().force_login(user, backend)

    def generic(self, *args, user=USER_IGNORE, **kwargs):
        """
        Force given user to login, then fetch request and return response.

        This adds the keyword argument `user` to all methods that end up calling `generic`, including
        `get()` and `post()`.

        When no `user` is set to `ExtendedClient.USER_IGNORE` (default), no special action is taken.
        When `user` is `None`, the request will be performed on an anonymous request (user logged out).
        When `user` is provided and not `None`, the request will be performed with that user logged in.
        """
        self.force_login(user)
        return super().generic(*args, **kwargs)



import re

from django.contrib.auth.models import User
from django.urls import reverse

from django_marina.test import ExtendedTestCase


class ExtendedTestCaseTestCase(ExtendedTestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.url_access_all = reverse("home")
        cls.url_access_authenticated = reverse("access_authenticated")
        cls.url_access_superuser = reverse("access_superuser")
        cls.url_does_not_exist = "/does-not-exist"
        cls.user = User(username="user", is_superuser=False)
        cls.user.save()
        cls.superuser = User(username="superuser", is_superuser=True)
        cls.superuser.save()

    def test_assert_response_status_code(self):
        response = self.client.get(self.url_access_all)
        self.assertResponseStatusCode(response, 200)
        with self.assertRaises(AssertionError):
            self.assertResponseStatusCode(response, "200")
        with self.assertRaises(AssertionError):
            self.assertResponseStatusCode(response, [200, 201])

    def test_assert_response_status_code_message(self):
        response = self.client.get(self.url_access_all)
        with self.assertRaisesMessage(AssertionError, "Invalid response code 200 (expected 100)."):
            self.assertResponseStatusCode(response, 100)

    def test_assert_response_ok(self):
        response = self.client.get(self.url_access_all)
        self.assertResponseOk(response)
        response = self.client.get(self.url_does_not_exist)
        with self.assertRaises(AssertionError):
            self.assertResponseOk(response)

    def test_assert_login_required(self):
        self.assertLoginRequired(self.url_access_authenticated)
        with self.assertRaises(AssertionError):
            self.assertLoginRequired(self.url_access_all)
        with self.assertRaises(AssertionError):
            self.assertLoginRequired(reverse("redirect-root"))

    def test_assert_login_not_required(self):
        self.assertLoginNotRequired(reverse("home"))
        with self.assertRaises(AssertionError):
            self.assertLoginNotRequired(self.url_access_authenticated)

    def test_methods(self):
        self.assertAllowed(self.url_access_all, user=None, method="get")
        self.assertAllowed(self.url_access_all, user=None, method="post")
        with self.assertRaises(AssertionError):
            self.assertAllowed(self.url_access_all, user=None, method="illegal")

    def test_assert_allowed(self):
        self.assertAllowed(self.url_access_all, user=None)
        self.assertAllowed(self.url_access_all, user=self.user)
        self.assertAllowed(self.url_access_all, user=self.superuser)
        self.assertAllowed(self.url_access_all, user=[None, self.user, self.superuser])

        self.assertAllowed(self.url_access_authenticated, user=self.user)
        self.assertAllowed(self.url_access_authenticated, user=self.superuser)
        with self.assertRaises(AssertionError):
            self.assertAllowed(self.url_access_authenticated, user=None)
        self.assertAllowed(self.url_access_all, user=[self.user, self.superuser])

        self.assertAllowed(self.url_access_superuser, user=self.superuser)
        with self.assertRaises(AssertionError):
            self.assertAllowed(self.url_access_superuser, user=None)
        with self.assertRaises(AssertionError):
            self.assertAllowed(self.url_access_superuser, user=self.user)

        self.assertAllowed(self.url_access_superuser, user=[self.superuser])
        with self.assertRaises(AssertionError):
            self.assertAllowed(self.url_access_superuser, user=[None, self.user, self.superuser])

    def test_assert_forbidden(self):
        with self.assertRaises(AssertionError):
            self.assertForbidden(self.url_access_all, user=None)
        with self.assertRaises(AssertionError):
            self.assertForbidden(self.url_access_all, user=self.user)
        with self.assertRaises(AssertionError):
            self.assertForbidden(self.url_access_all, user=self.superuser)

        with self.assertRaises(AssertionError):
            # Be careful with assertForbidden and user=None.
            # If login is required, the response returned may have a forbidden or redirect status code.
            # In this test, we will get a redirect code.
            self.assertForbidden(self.url_access_authenticated, user=None)
        with self.assertRaises(AssertionError):
            self.assertForbidden(self.url_access_authenticated, user=self.user)
        with self.assertRaises(AssertionError):
            self.assertForbidden(self.url_access_authenticated, user=self.superuser)

        with self.assertRaises(AssertionError):
            self.assertForbidden(self.url_access_superuser, user=None)
        self.assertForbidden(self.url_access_superuser, user=self.user)
        with self.assertRaises(AssertionError):
            self.assertForbidden(self.url_access_superuser, user=self.superuser)

    def test_assert_not_found(self):
        self.assertNotFound(self.url_does_not_exist, user=None)
        self.assertNotFound(self.url_does_not_exist, user=self.user)
        self.assertNotFound(self.url_does_not_exist, user=self.superuser)
        self.assertNotFound(self.url_does_not_exist, user=[None, self.user, self.superuser])

        with self.assertRaises(AssertionError):
            self.assertNotFound(self.url_access_all, user=None)
        with self.assertRaises(AssertionError):
            self.assertNotFound(self.url_access_all, user=self.user)
        with self.assertRaises(AssertionError):
            self.assertNotFound(self.url_access_all, user=self.superuser)

    def test_assert_contains_tag(self):
        response = self.client.get(self.url_access_all)
        self.assertContainsTag(response, "h1")
        with self.assertRaises(AssertionError):
            self.assertContainsTag(response, "h2")

    def test_assert_not_contains_tag(self):
        response = self.client.get(self.url_access_all)
        self.assertNotContainsTag(response, "h2")
        with self.assertRaises(AssertionError):
            self.assertNotContainsTag(response, "h1")

    def test_soup_message_prefix(self):
        response = self.client.get(self.url_access_all)
        with self.assertRaises(AssertionError):
            self.assertContainsSelector(response, "body h1 a", msg_prefix="MSG-PREFIX")

    def test_assert_contains_selector(self):
        response = self.client.get(self.url_access_all)
        self.assertContainsSelector(response, "body h1")
        self.assertContainsSelector(response, "body h1", count=1)
        with self.assertRaises(AssertionError):
            self.assertContainsSelector(response, "body h1", count=2)
        with self.assertRaises(AssertionError):
            self.assertContainsSelector(response, "body h1 a")

    def test_assert_not_contains_selector(self):
        response = self.client.get(self.url_access_all)
        self.assertNotContainsSelector(response, "body h1 a")
        with self.assertRaises(AssertionError):
            self.assertNotContainsSelector(response, "body h1")

    def test_assert_tag_with_string(self):
        response = self.client.get(self.url_access_all)
        self.assertNotContainsTag(response, "h1", string="Goodbye World!")
        self.assertContainsTag(response, "h1", string="Hello World!")
        self.assertNotContainsTag(response, "h1", string="Hello")
        self.assertContainsTag(response, "h1", string=re.compile("Hello"))

    def test_assert_selector_with_string(self):
        response = self.client.get(self.url_access_all)
        self.assertNotContainsSelector(response, "h1", string="Goodbye World!")
        self.assertContainsSelector(response, "h1", string="Hello World!")
        self.assertNotContainsSelector(response, "h1", string="Hello")
        self.assertContainsSelector(response, "h1", string=re.compile("Hello"))


class MessagesTestCase(ExtendedTestCase):
    def test_assert_has_message(self):
        message = "I have a message for you."
        partial_message = "message for you"
        url = reverse("message")

        response = self.client.get(url)
        with self.assertRaises(AssertionError):
            self.assertHasMessage(response, message)

        response = self.client.get(url, data={"message": message})
        self.assertHasMessage(response, message)
        with self.assertRaises(AssertionError):
            self.assertHasMessage(response, partial_message)


class HtmlTestCase(ExtendedTestCase):
    def test_assert_html_equal(self):
        html1 = '<span aria-foo="bar">foo-bar</span>'
        html2 = "<span>foo-bar</span>"
        with self.assertRaises(AssertionError):
            self.assertHTMLEqual(html1, html2)
        self.assertHTMLEqual(html1, html2, ignore_attrs=["aria-foo"])



from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext_lazy


class ProtectedModelMixin:
    """
    Add checks to update and delete methods of an instance.

    The use case is to protect against changes when update or delete is allowed for a user (superuser, admin or
    otherwise), but should not be executed because of objects that depend on this instance. In such a case, the foreign
    key relation to the model might benefit from `on_delete=models.PROTECT`.
    """

    # Default message shown to user when object has update protection.
    update_protection_message = gettext_lazy("This object has update protection.")

    # Default message shown to user when object has delete protection.
    delete_protection_message = gettext_lazy("This object has delete protection.")

    @property
    def has_update_protection(self):
        """Return whether this object has update protection."""
        return False

    def get_update_protection_message(self):
        """Return message shown to user when object has update protection."""
        return self.update_protection_message

    def save(self, *args, **kwargs):
        """Save the instance, unless it has a primary key and update protection."""
        if self.pk and self.has_update_protection:
            raise PermissionDenied(self.get_update_protection_message())
        return super().save(*args, **kwargs)

    @property
    def has_delete_protection(self):
        """Return whether this object has delete protection."""
        return False

    def get_delete_protection_message(self):
        """Return message shown to user when object has delete protection."""
        return self.delete_protection_message

    def delete(self, *args, **kwargs):
        """Delete the instance, unlessit has delete protection."""
        if self.has_delete_protection:
            raise PermissionDenied(self.get_delete_protection_message())
        return super().delete(*args, **kwargs)



from http import HTTPStatus

from django.contrib.messages import get_messages
from django.test import TestCase

from django_marina.html import remove_attrs

from .clients import ExtendedClient


def _login_url(next):
    """Return login url that Django uses in its redirect to login."""
    from django.contrib.auth.views import redirect_to_login

    redirect_response = redirect_to_login(next)
    return redirect_response["Location"]


def _msg_prefix_add(msg_prefix, value):
    """Return msg_prefix with added value."""
    msg_prefix = f"{msg_prefix}: " if msg_prefix else ""
    return f"{msg_prefix}{value}"


class ExtendedTestCase(TestCase):
    """TestCase with a extended client and extra features for asserting response content."""

    client_class = ExtendedClient

    def _response(self, path, user, **kwargs):
        """Return response on a separate client instance, so without any side effects."""
        client = self.client_class()
        method = kwargs.pop("method", "get").lower()
        if method == "get":
            response = client.get(path, user=user, **kwargs)
        elif method == "post":
            response = client.post(path, user=user, **kwargs)
        else:
            raise AssertionError(f"Invalid method {method} (expected GET or POST).")
        return response

    def assertResponseStatusCode(self, response, status_code, msg_prefix=None):
        """
        Assert that response has given status code.

        :param response: HttpResponse
        :param status_code: int
        :param msg_prefix: str
        """
        self.assertEqual(
            response.status_code,
            status_code,
            _msg_prefix_add(
                msg_prefix,
                f"Invalid response code {response.status_code} (expected {status_code}).",
            ),
        )

    def assertResponseOk(self, response):
        """
        Assert that response has status code OK.

        :param response: HttpResponse
        """
        self.assertResponseStatusCode(response, HTTPStatus.OK)

    def assertLoginRequired(self, path, **kwargs):
        """
        Make request while not logged in, assert that response has status code Forbidden or redirects to login page.

        :param path: Path for request
        :param kwargs: Kwargs for request
        """
        response = self._response(path, user=None, **kwargs)
        if response.status_code == HTTPStatus.FOUND:
            self.assertRedirects(response, _login_url(path))
        else:
            self.assertResponseStatusCode(response, HTTPStatus.FORBIDDEN)

    def _assertStatusCode(self, path, user, status_code, msg_prefix=None, **kwargs):
        """
        Make request, assert that response has given status code.

        :param path: Path for request
        :param user: User for request
        :param kwargs: Kwargs for request
        """
        users = user if isinstance(user, (list, tuple)) else [user]
        for user in users:
            response = self._response(path, user=user, **kwargs)
            self.assertResponseStatusCode(response, status_code, _msg_prefix_add(msg_prefix, user))

    def assertLoginNotRequired(self, path, **kwargs):
        """
        Make request while not logged in, assert that response has status code OK.

        :param path: Path for request
        :param kwargs: Kwargs for request
        """
        self._assertStatusCode(path, user=None, status_code=HTTPStatus.OK, **kwargs)

    def assertAllowed(self, path, user, **kwargs):
        """
        Make request, assert that response has status code OK.

        If `user` contains a list of users, the assertion will be made for every user in that list.

        :param path: Path for request
        :param user: User or list of users for request
        :param kwargs: Kwargs for request
        """
        self._assertStatusCode(path, user=user, status_code=HTTPStatus.OK, **kwargs)

    def assertForbidden(self, path, user, **kwargs):
        """
        Make request, assert that response has status code Forbidden.

        If `user` contains a list of users, the assertion will be made for every user in that list.

        :param path: Path for request
        :param user: User or list of users for request
        :param kwargs: Kwargs for request
        """
        self._assertStatusCode(path, user=user, status_code=HTTPStatus.FORBIDDEN, **kwargs)

    def assertNotFound(self, path, user, **kwargs):
        """
        Make request, assert that response has status code Not Found.

        If `user` contains a list of users, the assertion will be made for every user in that list.

        :param path: Path for request
        :param user: User or list of users for request
        :param kwargs: Kwargs for request
        """
        self._assertStatusCode(path, user=user, status_code=HTTPStatus.NOT_FOUND, **kwargs)

    def assertHasMessage(self, response, message):
        """
        Assert that response has given message.

        :param response: Response object
        :param message: Full text of message to check for
        """
        self.assertIn(message, [m.message for m in get_messages(response.wsgi_request)])

    def _get_soup_num_results(self, response, soup_method, soup_args, soup_kwargs):
        """Get number of results in response for BeautifulSoup query."""
        # Extract string kwarg for selectors since soup does not implement it
        string = soup_kwargs.pop("string", None) if soup_method == "select" else None

        method = getattr(response.soup, soup_method)
        results = method(*soup_args, **soup_kwargs)

        # Search with string kwarg for selectors, by applying it to found tags
        if string is not None:
            results = [result for result in results if result.find(string=string)]

        return len(results)

    def _assert_soup(
        self,
        response,
        soup_method,
        soup_args,
        soup_kwargs,
        status_code,
        count,
        msg_prefix,
    ):
        """Handle assertions that use BeautifulSoup, with interface similar to assertContains and assertNotContains."""
        self.assertEqual(
            response.status_code,
            status_code,
            _msg_prefix_add(
                msg_prefix,
                f"Couldn't retrieve content: Response code was {response.status_code} (expected {status_code}).",
            ),
        )

        soup_query = f"{soup_args} {soup_kwargs}"
        num_results = self._get_soup_num_results(response, soup_method, soup_args, soup_kwargs)

        if count == 0:
            self.assertEqual(
                num_results,
                0,
                _msg_prefix_add(msg_prefix, f"Response should not contain {soup_query}"),
            )
        elif count is None:
            self.assertTrue(
                num_results != 0,
                _msg_prefix_add(msg_prefix, f"Couldn't find {soup_query} in response"),
            )
        else:
            self.assertEqual(
                num_results,
                count,
                _msg_prefix_add(
                    msg_prefix,
                    f"Found {num_results} instances of {soup_query} in response (expected {count})",
                ),
            )

    def assertContainsTag(self, response, name=None, count=None, status_code=200, msg_prefix="", **kwargs):
        """
        Assert that tag can be found in response, using BeautifulSoup find notation.

        Documentation on tag finding: https://www.crummy.com/software/BeautifulSoup/bs4/doc/#kinds-of-filters
        """
        self._assert_soup(
            response,
            soup_method="find_all",
            soup_args=[name],
            soup_kwargs=kwargs,
            status_code=status_code,
            count=count,
            msg_prefix=msg_prefix,
        )

    def assertNotContainsTag(self, response, name=None, status_code=200, msg_prefix="", **kwargs):
        """
        Assert that tag cannot be found in response, using BeautifulSoup find notation.

        See `assertContainsTag` for documentation.
        """
        self._assert_soup(
            response,
            soup_method="find_all",
            soup_args=[name],
            soup_kwargs=kwargs,
            status_code=status_code,
            count=0,
            msg_prefix=msg_prefix,
        )

    def assertContainsSelector(self, response, selector, count=None, status_code=200, msg_prefix="", **kwargs):
        """
        Assert that CSS selector can be found in response.

        Documentation on CSS selectors: https://www.crummy.com/software/BeautifulSoup/bs4/doc/#css-selectors

        To search within the found selectors, you can optionally provide a `string` argument. This will filter
        the found tags. See https://www.crummy.com/software/BeautifulSoup/bs4/doc/#the-string-argument
        """
        self._assert_soup(
            response,
            soup_method="select",
            soup_args=[selector],
            soup_kwargs=kwargs,
            status_code=status_code,
            count=count,
            msg_prefix=msg_prefix,
        )

    def assertNotContainsSelector(self, response, selector, status_code=200, msg_prefix="", **kwargs):
        """Assert that CSS selector cannot be found in response."""
        self._assert_soup(
            response,
            soup_method="select",
            soup_args=[selector],
            soup_kwargs=kwargs,
            status_code=status_code,
            count=0,
            msg_prefix=msg_prefix,
        )

    def assertHTMLEqual(self, html1, html2, msg=None, ignore_attrs=None):
        """
        Assert that the strings html1 and html2 are equal, except for certain attributes to ignore.

        The comparison is based on HTML semantics.
        """
        return super().assertHTMLEqual(remove_attrs(html1, ignore_attrs), remove_attrs(html2, ignore_attrs), msg)



from bs4 import BeautifulSoup


def remove_attrs(html, attrs=None):
    """Return html string without the specified tags."""
    if html and attrs:
        soup = BeautifulSoup(html, "html.parser")
        soup = _remove_attrs(soup, attrs)
        html = f"{soup}"
    return html


def _remove_attrs(soup, attrs):
    for attr in attrs:
        for tag in soup.findAll(attrs={attr: True}):
            del tag.attrs[attr]
    return soup



from django.test import TestCase

from django_marina.html import remove_attrs


class HtmlTestCase(TestCase):
    def test_remove_attrs(self):
        self.assertEqual(remove_attrs('<span foo="bar"></span>', ["foo"]), "<span></span>")



from django.test import TestCase


class VersionTest(TestCase):
    """Test presence of package version."""

    def test_version(self):
        import django_marina

        version = django_marina.__version__
        version_parts = version.split(".")
        self.assertTrue(len(version_parts) in (2, 3))



from django.core.exceptions import PermissionDenied
from django.test import TestCase

from django_marina.db import DisableMigrations

from .models import ProtectedModel


class ProtectedModelMixinTestCase(TestCase):
    def test_update_protected(self):
        protected = ProtectedModel(name="protected", is_update_protected=True)
        protected.save()
        with self.assertRaises(PermissionDenied):
            protected.save()

    def test_update_unprotected(self):
        unprotected = ProtectedModel(name="protected", is_update_protected=False)
        unprotected.save()
        unprotected.name = "changed"
        unprotected.save()
        self.assertEqual(unprotected.name, "changed")

    def test_get_protected_against_update_message(self):
        protected = ProtectedModel(name="protected", is_update_protected=True)
        protected.save()
        message = protected.get_update_protection_message()
        self.assertEqual(message, "This object has update protection.")

    def test_delete_protected(self):
        protected = ProtectedModel(name="protected", is_delete_protected=True)
        protected.save()
        with self.assertRaises(PermissionDenied):
            protected.delete()

    def test_delete_unprotected(self):
        unprotected = ProtectedModel(name="protected", is_delete_protected=False)
        unprotected.save()
        unprotected.delete()
        self.assertIsNone(ProtectedModel.objects.filter(pk=unprotected.pk).first())

    def test_get_protected_against_delete_message(self):
        protected = ProtectedModel(name="protected", is_delete_protected=True)
        protected.save()
        message = protected.get_delete_protection_message()
        self.assertEqual(message, "This object has delete protection.")


class DisableMigrationsTestCase(TestCase):
    def test_disable_migrations(self):
        instance = DisableMigrations()
        self.assertTrue("anything" in instance)
        self.assertIsNone(instance["anything"])
